#ifndef __SPU_UTILS_H__
#define __SPU_UTILS_H__
#include <NestStep_spulib/neststep_spu.h>
#define NAME_OFFSET 1024
#define NAME_REDUCE 1
#define MAX_SPUS 8
#include "spu_timer.h"
#include "bl_common.h"

#define BUF_ALIGN 128

// this module contains functions to handle NestStep block distributed arrays (BArrF and BArrD).
//
// The basic usage is to loop over get_sw_BArrF(BArrF_Handler *h) until it returns zero.
// The handlers take care of double buffering, buffer alignment etc. See map.c for usage example

// __attribute__((aligned(16))) does not work for alignments larger than 16 bytes for stack allocated
// objects. 128 byte alignment is therefor found inside a slightly larger than needed buffer.
// (128 byte alignment is necessary to get more than ~1/2 the available memory bandwidth).

// get and store that can handle sizes that is not a factor of 16b
void get_BArrF_dA(BlockDistArray* _barr, float* lsaddr, int lbound, int ubound);
void store_BArrF_dA(BlockDistArray* _barr, float* lsaddr, int lbound, int ubound);

void get_BArrD_dA(BlockDistArray* _barr, double* lsaddr, int lbound, int ubound);
void store_BArrD_dA(BlockDistArray* _barr, double* lsaddr, int lbound, int ubound);


typedef struct BArrF_Handler_ {
	float buf1[CHUNKSIZE_F+BUF_ALIGN/4]__attribute__((aligned(16)));
	float buf2[CHUNKSIZE_F+BUF_ALIGN/4]__attribute__((aligned(16)));
	BlockDistArray* a;
	int counter;
	int blockSize;
	int start;
	int part;
	float *next;
	int nextSize;
	int nextStart;
	
	float *current;
	int currentSize;
	int currentStart;
} BArrF_Handler;

// init handlers (bs = block size/chunk size (<=16 kiB, factor of 128 bytes))
// init for read handler
void init_BArrF_Get(BArrF_Handler *h, BlockDistArray* a, int bs);
// init for write handler
void init_BArrF_Put(BArrF_Handler *h, BlockDistArray* a, int bs);
// get next and swap buffers for read handler
int get_sw_BArrF(BArrF_Handler *h);
// get next and swap buffers for write handler
int put_sw_BArrF(BArrF_Handler *h);
// wait for last put chunk transfer to complete
void wait_put_BArrF_done(BArrF_Handler *h);

typedef struct BArrD_Handler_ {
	double buf1[CHUNKSIZE_D+BUF_ALIGN/8] __attribute__((aligned(16)));
	double buf2[CHUNKSIZE_D+BUF_ALIGN/8] __attribute__((aligned(16)));
	BlockDistArray* a;
	int counter;
	int blockSize;
	int start;
	int part;
	double *next;
	int nextSize;
	int nextStart;
	double *current;
	int currentSize;
	int currentStart;
} BArrD_Handler;


void init_BArrD_Get(BArrD_Handler *h, BlockDistArray* a, int bs);
void init_BArrD_Put(BArrD_Handler *h, BlockDistArray* a, int bs);
int get_sw_BArrD(BArrD_Handler *h);
int put_sw_BArrD(BArrD_Handler *h);
void wait_put_BArrD_done(BArrD_Handler *h);
void wait_put_done(BArrF_Handler *h);

void printBArrF(BlockDistArray* x, int n, int nMax);

void printBArrD(BlockDistArray* x, int n, int nMax);


void get_PArrF_dA(PrivateArray* _parr, float* lsaddr, int lbound, int ubound);
void store_PArrF_dA(PrivateArray* _parr, float* lsaddr, int lbound, int ubound);

void get_PArrD_dA(PrivateArray* _parr, double* lsaddr, int lbound, int ubound);
void store_PArrD_dA(PrivateArray* _parr, double* lsaddr, int lbound, int ubound);


typedef struct PArrF_Handler_ {
	float buf1[CHUNKSIZE_F+BUF_ALIGN/4]__attribute__((aligned(16)));
	float buf2[CHUNKSIZE_F+BUF_ALIGN/4]__attribute__((aligned(16)));
	PrivateArray* a;
	int counter;
	int blockSize;
	int size;
	float *next;
	int nextSize;
	int nextStart;
	
	float *current;
	int currentSize;
	int currentStart;
} PArrF_Handler;

void init_PArrF_Get(PArrF_Handler *h, PrivateArray* a, int bs);
void init_PArrF_Put(PArrF_Handler *h, PrivateArray* a, int bs);
int get_sw_PArrF(PArrF_Handler *h);
int put_sw_PArrF(PArrF_Handler *h);
void wait_put_PArrF_done(PArrF_Handler *h);

typedef struct PArrD_Handler_ {
	double buf1[CHUNKSIZE_D+BUF_ALIGN/8] __attribute__((aligned(16)));
	double buf2[CHUNKSIZE_D+BUF_ALIGN/8] __attribute__((aligned(16)));
	PrivateArray* a;
	int counter;
	int blockSize;
	int size;
	double *next;
	int nextSize;
	int nextStart;
	double *current;
	int currentSize;
	int currentStart;
} PArrD_Handler;

void init_PArrD_Get(PArrD_Handler *h, PrivateArray* a, int bs);
void init_PArrD_Put(PArrD_Handler *h, PrivateArray* a, int bs);
int get_sw_PArrD(PArrD_Handler *h);
int put_sw_PArrD(PArrD_Handler *h);
void wait_put_PArrD_done(PArrD_Handler *h);


static inline int getBDALocalSize(BlockDistArray* x)
{
	return x->localsize;
};

static inline int getBDALowerBound(BlockDistArray* x)
{
	return x->lower;
};

static inline int getPASize(PrivateArray* x)
{
	return x->size;
};


static inline int dmaPad(int size)
{
	if(size %4)
		size += 4-size%4;
	return size;
}

void *fix_align(unsigned int align, void *ptr);


#endif
